import time
from urllib.parse import urljoin

import requests
from requests import Response, HTTPError

from biolib.biolib_logging import logger_no_user_data
from biolib.typing_utils import Dict, Optional, Union
from biolib.biolib_api_client import BiolibApiClient as DeprecatedApiClient

OptionalHeaders = Optional[Dict[str, Union[str, None]]]


class ApiClient:

    def __init__(self):
        self._session = requests.session()

    def get(
            self,
            url: str,
            params: Optional[Dict[str, Union[str, int]]] = None,
            headers: OptionalHeaders = None,
            authenticate: bool = True,
    ) -> Response:
        retries = 10
        for retry_count in range(retries):
            if retry_count > 0:
                time.sleep(5 * retry_count)
                logger_no_user_data.debug('Retrying HTTP GET request...')
            try:
                response: Response = self._session.get(
                    headers=self._get_headers(headers, authenticate),
                    params=params,
                    timeout=60,
                    url=self._get_absolute_url(url),
                )
                if response.status_code == 502:
                    logger_no_user_data.debug(f'HTTP GET request failed with status 502 for "{url}"')
                    continue

                ApiClient.raise_for_status(response)
                return response
            except requests.exceptions.ReadTimeout:
                logger_no_user_data.debug(f'HTTP GET request failed with read timeout for "{url}"')
                continue

        raise Exception(f'HTTP GET request failed after {retries} retries for "{url}"')

    def post(self, path: str, data: Union[Dict, bytes], headers: OptionalHeaders = None) -> Response:
        retries = 3
        for retry_count in range(retries):
            if retry_count > 0:
                time.sleep(5 * retry_count)
                logger_no_user_data.debug('Retrying HTTP POST request...')
            try:
                response: Response = self._session.post(
                    headers=self._get_headers(headers),
                    data=data if not isinstance(data, dict) else None,
                    json=data if isinstance(data, dict) else None,
                    timeout=10 if isinstance(data, dict) else 180,  # TODO: Calculate timeout based on data size
                    url=self._get_absolute_url(path),
                )
                if response.status_code == 502:
                    logger_no_user_data.debug(f'HTTP POST request failed with status 502 for "{path}"')
                    continue

                ApiClient.raise_for_status(response)
                return response
            except requests.exceptions.ReadTimeout:
                logger_no_user_data.debug(f'HTTP POST request failed with read timeout for "{path}"')
                continue

        raise Exception(f'HTTP POST request failed after {retries} retries for "{path}"')

    def patch(self, path: str, data: Dict, headers: OptionalHeaders = None) -> Response:
        response: Response = self._session.patch(
            headers=self._get_headers(headers),
            json=data,
            timeout=10,
            url=self._get_absolute_url(path),
        )
        ApiClient.raise_for_status(response)
        return response

    @staticmethod
    def raise_for_status(response):
        # Logic taken from `requests.Response.raise_for_status()`
        http_error_msg = ''
        reason = response.text
        if 400 <= response.status_code < 500:
            http_error_msg = u'%s Client Error: %s for url: %s' % (response.status_code, reason, response.url)

        elif 500 <= response.status_code < 600:
            http_error_msg = u'%s Server Error: %s for url: %s' % (response.status_code, reason, response.url)

        if http_error_msg:
            raise HTTPError(http_error_msg, response=response)

    @staticmethod
    def _get_headers(opt_headers: OptionalHeaders = None, authenticate: bool = True) -> Dict[str, str]:
        # Only keep header keys with a value
        headers: Dict[str, str] = {key: value for key, value in (opt_headers or {}).items() if value}

        deprecated_api_client = DeprecatedApiClient.get()

        if deprecated_api_client.is_signed_in:
            deprecated_api_client.refresh_access_token()

        # Adding access_token outside is_signed_in check as job_worker.py currently sets access_token
        # without setting refresh_token
        access_token = deprecated_api_client.access_token
        if access_token and authenticate:
            headers['Authorization'] = f'Bearer {access_token}'

        return headers

    @staticmethod
    def _get_absolute_url(path: str) -> str:
        deprecated_api_client = DeprecatedApiClient.get()
        base_api_url = urljoin(deprecated_api_client.base_url, '/api/')
        return urljoin(base_api_url, path.strip('/') + '/')



import threading
import time
import subprocess
from datetime import datetime

import docker.errors  # type: ignore
from docker.models.containers import Container  # type: ignore

import biolib.api.client
from biolib.biolib_logging import logger_no_user_data
from biolib.typing_utils import List, TypedDict, Optional, Dict, cast


class UtilizationMetricSample(TypedDict):
    cpu_usage_in_percent: float
    gpu_usage_in_percent: Optional[float]
    memory_usage_in_percent: float


class AggregatedUtilizationMetrics(TypedDict):
    cpu_average_usage_in_percent: float
    cpu_max_usage_in_percent: float
    gpu_average_usage_in_percent: Optional[float]
    gpu_max_usage_in_percent: Optional[float]
    memory_average_usage_in_percent: float
    memory_max_usage_in_percent: float
    recorded_at: str
    sampling_period_in_milliseconds: int


class UtilizationReporterThread(threading.Thread):
    def __init__(self, container: Container, job_uuid: str, compute_node_auth_token: str):
        super().__init__(daemon=True)
        self._container_object: Container = container
        self._job_uuid: str = job_uuid
        self._compute_node_auth_token: str = compute_node_auth_token

        self._sampling_period_in_milliseconds = 1_000
        self._samples_between_writes = 60

    def run(self) -> None:
        logger_no_user_data.debug(f'Job "{self._job_uuid}" utilization metrics reporter thread started')
        prev_cpu_usage: Optional[float] = None
        prev_cpu_system_usage: Optional[float] = None
        metric_samples: List[UtilizationMetricSample] = []
        while True:
            stats = self._get_container_stats()
            if not stats:
                break

            if prev_cpu_usage is None or prev_cpu_system_usage is None:
                prev_cpu_usage = stats['cpu_stats']['cpu_usage']['total_usage']
                prev_cpu_system_usage = stats['cpu_stats']['system_cpu_usage']
                continue

            # Calculate CPU usage
            cpu_usage_delta_ns = stats['cpu_stats']['cpu_usage']['total_usage'] - prev_cpu_usage
            cpu_system_usage_delta_ns = stats['cpu_stats']['system_cpu_usage'] - prev_cpu_system_usage
            cpu_usage_in_percent = round((cpu_usage_delta_ns / cpu_system_usage_delta_ns) * 100, ndigits=2)

            # Set previous usage
            prev_cpu_usage = stats['cpu_stats']['cpu_usage']['total_usage']
            prev_cpu_system_usage = stats['cpu_stats']['system_cpu_usage']

            # Calculate Memory usage
            memory_usage_in_percent = round(
                stats['memory_stats']['usage'] / stats['memory_stats']['limit'] * 100,
                ndigits=2,
            )

            metric_samples.append(UtilizationMetricSample(
                cpu_usage_in_percent=cpu_usage_in_percent,
                memory_usage_in_percent=memory_usage_in_percent,
                gpu_usage_in_percent=self._get_gpu_utilization_in_percent(),
            ))

            if len(metric_samples) >= self._samples_between_writes:
                self._report_aggregated_utilization_metric(metric_samples)
                metric_samples = []

            time.sleep(secs=self._sampling_period_in_milliseconds / 1_000)

        # Write the remaining samples after container has exited
        self._report_aggregated_utilization_metric(metric_samples)
        logger_no_user_data.debug(f'Job "{self._job_uuid}" utilization metrics reporter thread exiting')

    @staticmethod
    def _get_gpu_utilization_in_percent() -> Optional[float]:
        try:
            cmd = 'nvidia-smi --query-gpu=utilization.gpu --format=csv,noheader'
            utilization = subprocess.check_output(cmd, shell=True, stderr=subprocess.DEVNULL).decode('utf-8')
            utilization_for_each_gpu = [float(x.replace(' %', '')) for x in utilization.strip().split('\n')]
            utilization_for_first_gpu = utilization_for_each_gpu[0]
            return utilization_for_first_gpu
        except BaseException as error:
            logger_no_user_data.exception(f'Failed to get GPU utilization got error: {error}')
            return None

    def _get_container_stats(self) -> Optional[Dict]:
        try:
            return cast(Dict, self._container_object.stats(stream=False))
        except docker.errors.NotFound:
            return None

    def _get_aggregated_utilization_metric_from_metric_samples(
            self,
            metric_samples: List[UtilizationMetricSample],
    ) -> AggregatedUtilizationMetrics:
        cpu_max_usage_in_percent: float = 0.0
        cpu_usage_in_percent_sum: float = 0.0
        gpu_max_usage_in_percent: Optional[float] = None
        gpu_usage_in_percent_sum: Optional[float] = None
        memory_max_usage_in_percent: float = 0.0
        memory_usage_in_percent_sum: float = 0.0

        for metric_sample in metric_samples:
            cpu_max_usage_in_percent = max(cpu_max_usage_in_percent, metric_sample['cpu_usage_in_percent'])
            cpu_usage_in_percent_sum += metric_sample['cpu_usage_in_percent']
            memory_max_usage_in_percent = max(memory_max_usage_in_percent, metric_sample['memory_usage_in_percent'])
            memory_usage_in_percent_sum += metric_sample['memory_usage_in_percent']

            if metric_sample['gpu_usage_in_percent'] is not None:
                if gpu_max_usage_in_percent is None:
                    gpu_max_usage_in_percent = 0.0
                if gpu_usage_in_percent_sum is None:
                    gpu_usage_in_percent_sum = 0.0

                gpu_max_usage_in_percent = max(gpu_max_usage_in_percent, metric_sample['gpu_usage_in_percent'])
                gpu_usage_in_percent_sum += metric_sample['gpu_usage_in_percent']

        cpu_average_usage_in_percent = cpu_usage_in_percent_sum / len(metric_samples)
        memory_average_usage_in_percent = memory_usage_in_percent_sum / len(metric_samples)
        gpu_average_usage_in_percent = gpu_usage_in_percent_sum / len(metric_samples) \
            if gpu_usage_in_percent_sum is not None else None

        return AggregatedUtilizationMetrics(
            cpu_average_usage_in_percent=cpu_average_usage_in_percent,
            cpu_max_usage_in_percent=cpu_max_usage_in_percent,
            gpu_average_usage_in_percent=gpu_average_usage_in_percent,
            gpu_max_usage_in_percent=gpu_max_usage_in_percent,
            memory_average_usage_in_percent=memory_average_usage_in_percent,
            memory_max_usage_in_percent=memory_max_usage_in_percent,
            recorded_at=datetime.utcnow().isoformat(),
            sampling_period_in_milliseconds=self._sampling_period_in_milliseconds * self._samples_between_writes,
        )

    def _report_aggregated_utilization_metric(self, metric_samples: List[UtilizationMetricSample]) -> None:
        if len(metric_samples) == 0:
            logger_no_user_data.debug(f'Job "{self._job_uuid}" no metric samples to aggregate. Skipping reporting.')
            return

        aggregated_metrics = self._get_aggregated_utilization_metric_from_metric_samples(metric_samples)
        logger_no_user_data.debug(f'Job "{self._job_uuid}" reporting aggregated metrics {aggregated_metrics}')

        response = biolib.api.client.post(
            path=f'/internal/compute-nodes/jobs/{self._job_uuid}/utilization-metrics/',
            headers={'Compute-Node-Auth-Token': self._compute_node_auth_token},
            data=cast(Dict, aggregated_metrics),
        )

        if not response.ok:
            logger_no_user_data.error(
                f'Failed to report metrics got status {response.status_code} and error: {response.text}'
            )



import re
import os
import subprocess

import requests

import biolib.api
from biolib import biolib_errors
from biolib.typing_utils import Optional
from biolib.biolib_api_client.auth import BearerAuth
from biolib.biolib_api_client import BiolibApiClient, AppGetResponse
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger


def _get_git_branch_name() -> str:
    try:
        github_actions_branch_name = os.getenv('GITHUB_REF_NAME')
        if github_actions_branch_name:
            return github_actions_branch_name

        gitlab_ci_branch_name = os.getenv('CI_COMMIT_REF_NAME')
        if gitlab_ci_branch_name:
            return gitlab_ci_branch_name

        result = subprocess.run(['git', 'branch', '--show-current'], check=True, stdout=subprocess.PIPE, text=True)
        return result.stdout.strip()
    except BaseException:
        return ''


def _get_git_repository_url() -> str:
    try:
        result = subprocess.run(['git', 'remote', 'get-url', 'origin'], check=True, stdout=subprocess.PIPE, text=True)
        return result.stdout.strip()
    except BaseException:
        return ''


class BiolibAppApi:

    @staticmethod
    def get_by_uri(uri: str) -> AppGetResponse:
        # Replace protocol with @ (if supplied)
        uri = re.sub(r'^http(s)?://', '@', uri)
        # Replace frontend version path with app_uri compatible version (if supplied)
        uri = re.sub(r'/version/(?P<version>(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*))(/)?$', r':\g<version>', uri)

        try:
            response = biolib.api._client.get(url='/app/', params={'uri': uri})  # pylint: disable=protected-access
            app_response: AppGetResponse = response.json()
            return app_response

        except requests.exceptions.HTTPError as error:
            if error.response.status_code == 404:
                raise biolib_errors.NotFound(f'Application {uri} not found.') from None

            if error.response.status_code == 400:
                raise biolib_errors.BioLibError(error.response.content.decode()) from None

            raise error

    @staticmethod
    def push_app_version(
            app_id,
            zip_binary,
            author,
            app_name,
            set_as_active,
            app_version_id_to_copy_images_from: Optional[str],
    ):
        response = requests.post(
            f'{BiolibApiClient.get().base_url}/api/app_versions/',
            files={
                'source_files_zip': zip_binary,
            },
            data={
                'app': app_id,
                'set_as_active': 'true' if set_as_active else 'false',
                'state': 'published',
                'app_version_id_to_copy_images_from': app_version_id_to_copy_images_from,
                'git_branch_name': _get_git_branch_name(),
                'git_repository_url': _get_git_repository_url(),
            },
            auth=BearerAuth(BiolibApiClient.get().access_token)
        )
        if not response.ok:
            logger.error(f'Push failed for {author}/{app_name}:')
            raise BioLibError(response.text)

        # TODO: When response includes the version number, print the URL for the new app version
        logger.info(f'Initialized new app version for {author}/{app_name}.')
        return response.json()

    @staticmethod
    def update_app_version(app_version_id, data):
        response = requests.patch(
            f'{BiolibApiClient.get().base_url}/api/app_versions/{app_version_id}/',
            json=data,
            auth=BearerAuth(BiolibApiClient.get().access_token)
        )
        if not response.ok:
            logger.error(f'Failed to update app version {app_version_id}')
            raise BioLibError(response.text)

        return response.json()



from biolib.typing_utils import TypedDict


class LargeFileSystemVersion(TypedDict):
    presigned_download_url: str
    size_bytes: int
    uri: str
    uuid: str


class LargeFileSystem(TypedDict):
    uri: str
    uuid: str



from abc import ABC, abstractmethod
import io
import time

import requests
from requests import Response

from biolib.biolib_logging import logger
from biolib.typing_utils import Optional, Dict


class RemoteEndpoint(ABC):
    @abstractmethod
    def get_remote_url(self):
        pass


class IndexableBuffer(ABC):

    def __init__(self):
        self.pointer = 0

    @abstractmethod
    def get_data(self, start: int, length: int) -> bytes:
        pass

    def get_data_as_string(self, start: int, length: int) -> str:
        return self.get_data(start=start, length=length).decode()

    def get_data_as_int(self, start: int, length: int) -> int:
        return int.from_bytes(bytes=self.get_data(start=start, length=length), byteorder='big')

    def get_data_with_pointer(self, length: int) -> bytes:
        data = self.get_data(start=self.pointer, length=length)
        self.pointer += length
        return data

    def get_data_with_pointer_as_int(self, length: int) -> int:
        data = self.get_data_as_int(start=self.pointer, length=length)
        self.pointer += length
        return data

    def get_data_with_pointer_as_string(self, length: int) -> str:
        data = self.get_data_as_string(start=self.pointer, length=length)
        self.pointer += length
        return data


def _get_request_with_retries_for_timeout(
        url: str,
        headers: Optional[Dict[str, str]] = None,
        stream: bool = False,
) -> Response:
    retries = 10
    for retry_count in range(retries):
        if retry_count > 0:
            time.sleep(5 * retry_count)
            logger.debug('Retrying HTTP GET request...')

        try:
            response = requests.get(
                headers=headers,
                stream=stream,
                timeout=120,
                url=url,
            )
        except requests.exceptions.ReadTimeout:
            logger.debug('Got read timeout of HTTP GET request')
            continue

        if response.status_code == 502:
            logger.debug('Got 502 status for HTTP GET')
            continue

        if response.status_code == 400:
            logger.error(
                f'GET request for "{url}" got response with status {response.status_code} and '
                f'content: {response.content.decode()}'
            )

        if not response.ok:
            raise Exception(f'GET request for "{url}" got response with status {response.status_code}')

        return response

    raise Exception(f'Failed to complete HTTP GET request for "{url}" after {retries} retries')


class LocalFileIndexableBuffer(IndexableBuffer):

    def __init__(self, filename: str):
        super().__init__()
        self._filehandle = open(filename, 'rb')

    def get_data(self, start: int, length: int) -> bytes:
        if length < 0:
            raise Exception('get_data length must be positive')

        if length == 0:
            return bytes(0)

        self._filehandle.seek(start)
        data: bytes = self._filehandle.read(length)

        if len(data) != length:
            raise Exception(f'get_data got response of unexpected length. Got {len(data)} expected {length}.')

        return data


class RemoteIndexableBuffer(IndexableBuffer):

    def __init__(self, endpoint: RemoteEndpoint):
        super().__init__()
        self._endpoint = endpoint

    def get_data(self, start: int, length: int) -> bytes:
        if length < 0:
            raise Exception('get_data length must be positive')

        if length == 0:
            return bytes(0)

        end = start + length - 1
        response = _get_request_with_retries_for_timeout(
            url=self._endpoint.get_remote_url(),
            headers={'range': f'bytes={start}-{end}'},
        )

        data: bytes = response.content
        if len(data) != length:
            raise Exception(f'get_data got response of unexpected length. Got {len(data)} expected {length}.')

        return data


class InMemoryIndexableBuffer(IndexableBuffer):

    def __init__(self, data: bytes):
        super().__init__()
        self._buffer = data
        self._length_bytes = len(data)

    def get_data(self, start: int, length: int) -> bytes:
        end = start + length
        return self._buffer[start:end]

    def __len__(self):
        return self._length_bytes


class LazyLoadedFile:

    def __init__(self, path: str, buffer: IndexableBuffer, start: int, length: int):
        self._path = path
        self._buffer = buffer
        self._start = start
        self._length = length

    def __repr__(self) -> str:
        return f'File "{self._path}" with size of {self._length} bytes'

    @property
    def path(self) -> str:
        return self._path

    @property
    def start(self) -> int:
        return self._start

    @property
    def length(self) -> int:
        return self._length

    def get_file_handle(self) -> io.BufferedIOBase:
        return io.BytesIO(self.get_data())

    def get_data(self) -> bytes:
        return self._buffer.get_data(start=self._start, length=self._length)
